#include "RList.h"
#include <stdlib.h>

RList::RList(int itemsPerBlock = 20) //:
  // fPhysicalSize(itemsPerBlock),
  // fBlockSize(itemsPerBlock),
  // fItemCount(0)
{
    if (itemsPerBlock <= 0)
       itemsPerBlock = 1;
    fPhysicalSize =itemsPerBlock;
    fBlockSize = itemsPerBlock;
    fItemCount = 0;
    fObjectList = new (void*)[itemsPerBlock];
}

			RList::RList(const RList& anotherList) :
				  fPhysicalSize(anotherList.fPhysicalSize),
				  fBlockSize(anotherList.fBlockSize),
				  fItemCount(anotherList.fItemCount)
{
    fObjectList = new (void*)[fPhysicalSize];

    for(int i = 0; i < fItemCount; ++i)
    {
       fObjectList[i] = anotherList.fObjectList[i];
    }
}

	        RList::~RList()
{
    delete fObjectList;
}

RList	        &RList::operator=(const RList &from)
{
    fPhysicalSize = from.fPhysicalSize;
    fBlockSize = from.fBlockSize;
    fItemCount = from.fItemCount;

    delete fObjectList;
    fObjectList = new (void*)[fPhysicalSize];

    for(int i = 0; i < fItemCount; ++i)
    {
       fObjectList[i] = from.fObjectList[i];
    }    
    return *this;
}

/* Adding and removing items. */
bool	         RList::AddItem(void *item)
{
    Resize(fItemCount + 1);
    fObjectList[fItemCount] = item;
    fItemCount++;
    return true;
}

bool	         RList::AddItem(void *item, int atIndex)
{
    if ((atIndex >= 0) && (atIndex < fItemCount))
    {
      Resize(fItemCount + 1);
      for(int i = fItemCount; i > atIndex ; --i)
        fObjectList[i] = fObjectList[i - 1];
      fObjectList[atIndex] = item;
      fItemCount++;
      return true;
    }
    else
      return false;
}

bool	         RList::AddList(RList *newItems)
{
    Resize(fItemCount + newItems->CountItems());
    for(int i = 0; i < newItems->CountItems(); ++i)
       fObjectList[fItemCount + i] = newItems->ItemAt(i);
    fItemCount += newItems->CountItems();
    return true;
}

bool	         RList::AddList(RList *newItems, int atIndex)
{
    if ((atIndex >= 0) && (atIndex < fItemCount))
    {
      Resize(fItemCount + newItems->CountItems());
      void** tmp = new (void*)[fItemCount + newItems->CountItems()];
      for (int i = 0; i < atIndex; ++i)
         tmp[i] = fObjectList[i];
      for (int i = 0; i < newItems->CountItems(); ++i)
         tmp[atIndex + i] = newItems->ItemAt(i);
      for (int i = 0; i < fItemCount - atIndex; ++i)
         tmp[atIndex + newItems->CountItems() + i] = fObjectList[atIndex + i];
      for (int i = 0; i < fItemCount + newItems->CountItems(); ++i)
	     fObjectList[i] = tmp[i];
	  delete tmp;
      fItemCount += newItems->CountItems();
      return true;
    }
    else
      return false;
}

bool	         RList::RemoveItem(void *item)
{
    int          target = -1;
    
    for (int i = 0; (i < fItemCount) && (target == -1); ++i)
       if (fObjectList[i] == item)
          target = i;
    if (target != -1)
    {
       RemoveItem(target);
       return true;
    }
    else
       return false;
}

void	         *RList::RemoveItem(int index)
{
    if ((index >= 0) && (index < fItemCount))
    {
       void    *keep = fObjectList[index];
       
       for (int i = index; i < (fItemCount - 1); ++i)
          fObjectList[i] = fObjectList[i + 1];
       fItemCount--;    
       Resize(fItemCount);
       return keep;
    }
    else
       return NULL;
}

bool	         RList::RemoveItems(int index, int count)
{
    if ((index >= 0) && (count >= 0) && ((index + count) < fItemCount))
    {
       for (int i = index + count; i < fItemCount; ++i)
          fObjectList[i - count] = fObjectList[i];
       fItemCount -= count;
       Resize(fItemCount);
       return true;
    }
    else
       return false;
}

bool	         RList::ReplaceItem(int index, void *newItem)
{
    if ((index >= 0) && (index < fItemCount))
    {
       fObjectList[index] = newItem;
       return true;
    }
    return false;
}

void	         RList::MakeEmpty()
{
    delete fObjectList;
    fObjectList = new (void*)[1]; //to avoid deleting bad pointer
    fItemCount = 0;
    fPhysicalSize = 0;
}

/* Reordering items. */
void	         RList::SortItems(int (*cmp)(const void *, const void *))
{
    for (int i = 0; i < fItemCount; ++i)
       for (int j = 1; j < fItemCount - i; ++j)
       {
           if (cmp(fObjectList[j - 1], fObjectList[j]) > 0)
               SwapItems(j - 1, j);
       }
}

bool	         RList::SwapItems(int indexA, int indexB)
{
    void         *keep;
    
    if ((indexA >= 0) && (indexB >= 0)
        && (indexA < fItemCount) && (indexB < fItemCount))
    {
     keep = fObjectList[indexA];
     fObjectList[indexA] = fObjectList[indexB];
     fObjectList[indexB] = keep;
     return true;
    }
    else
     return false;
}

bool	         RList::MoveItem(int fromIndex, int toIndex)
{
    if ((fromIndex >= 0) && (fromIndex < fItemCount) &&
        (toIndex >= 0) && (toIndex < fItemCount))
    {
       void    *keep = fObjectList[fromIndex];
       for (unsigned int i = fromIndex; i < (unsigned int)fItemCount - 1; ++i)
          fObjectList[i] = fObjectList[i + 1];
       for (unsigned int i = (unsigned int)fItemCount - 1; i > (unsigned int)toIndex; --i)
          fObjectList[i] = fObjectList[i - 1];
       fObjectList[toIndex] = keep;
       return true;
	}
	return false;
}

/* Retrieving items. */
void	         *RList::ItemAt(int index) const
{
    return (ItemAtFast(index));
}

void	         *RList::ItemAtFast(int index) const
{
    if ((index >= 0) && (index < fItemCount))
       return fObjectList[index];
    else
       return NULL;
}

void	         *RList::FirstItem() const
{
    if (fItemCount > 0)
       return fObjectList[0];
    else
       return NULL;
}

void	         *RList::LastItem() const
{
    if (fItemCount > 0)
       return fObjectList[fItemCount - 1];
    else
       return NULL;
}

void	         *RList::Items() const
{
    return fObjectList;
}

/* Querying the list. */
bool	         RList::HasItem(void *item) const
{
   for (unsigned int i = 0; i < (unsigned int)fItemCount; ++i)
      if (fObjectList[i] == item)
         return true;
   return false;
}

int	         RList::IndexOf(void *item) const
{
    for (unsigned int i = 0; i < (unsigned int)fItemCount; ++i)
       if (fObjectList[i] == item)
          return i;
    return -1;
}

int	         RList::CountItems() const
{
    return fItemCount;
}

bool	         RList::IsEmpty() const
{
    if (fItemCount == 0)
       return true;
    else
       return false;   
}

/* Iterating over the list. */
void	         RList::DoForEach(bool (*func)(void *))
{
    void    **ptr = fObjectList;
    for (int i = 0; i < fItemCount; ++i)
       if (func(*ptr++) == true)
          break;
}

void	RList::DoForEach(bool (*func)(void *, void *), void * arg2)
{
    void    **ptr = fObjectList;
      for (int i = 0; i < fItemCount; ++i)
         if (func(*ptr++, arg2) == true)
            break;
}

void	RList::Resize(int count)
{
    if (count <= 0)
       return;
    if (count > fPhysicalSize)
    {
       while(count > fPhysicalSize)
          fPhysicalSize += fBlockSize;
       void**	newObjectList = new (void*)[fPhysicalSize];
       for(int i = 0; i < fItemCount; ++i)
          newObjectList[i] = fObjectList[i];
       delete fObjectList;
       fObjectList = newObjectList;
    }
    else
       if (count <= (fPhysicalSize - fBlockSize))
       {
          while(count <= (fPhysicalSize - fBlockSize))
          fPhysicalSize -= fBlockSize;
          void**	newObjectList = new (void*)[fPhysicalSize];
          for(int i = 0; (i < fItemCount) && (i < count); ++i)
             newObjectList[i] = fObjectList[i];
          delete fObjectList;
          fObjectList = newObjectList;
       }
    return;
}

